import os
import socket
import subprocess
import sys
import threading
from logging import getLogger
from pathlib import Path
from typing import Literal
import platform
import webbrowser
from argparse import ArgumentParser
from uvicorn import Config, Server

DEFAULT_METHOD_FOR_STARTING_SERVER = "same_terminal"
if platform.system() == "Darwin":
    DEFAULT_METHOD_FOR_STARTING_SERVER = "same_terminal"

if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/") and platform.system() == "Windows":
    DEFAULT_METHOD_FOR_STARTING_SERVER = "new_terminal"


logger = getLogger(__name__)

def is_port_in_use(port: int) -> bool:
    """Check if a port is already in use."""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(('127.0.0.1', port)) == 0
    

def start_server(port: int = 8000, method: Literal["new_terminal", "same_terminal"] = DEFAULT_METHOD_FOR_STARTING_SERVER) -> None:
    os_type = platform.system()

    if os_type == "Darwin" and method == "new_terminal":
        method = "same_terminal"
        print("Only same_terminal method is supported on macOS. Starting server in the same terminal.")
    print("Start server called with method: ", method)
    if is_port_in_use(port):
        return
    if method == "new_terminal":
        _start_server_in_new_terminal(port)
    elif method == "same_terminal":
        _start_server_in_same_terminal(port)
    else:
        print(f"Invalid method: {method}")


def _start_server_in_new_terminal(port: int = 8000):
    """Starts the server in a new terminal."""
    path_uvicorn="uvicorn"
    if os.name == "darwin":
        raise NotImplementedError("Starting a new terminal is not supported on macOS.")
    if os.name == "nt":
        path_exec = Path(sys.executable)
        path_uvicorn1 = path_exec.parent / "uvicorn.exe" # Henri    
        path_uvicorn2 = path_exec.parent / "Scripts" / "uvicorn.exe" # Lucas

        if not path_uvicorn1.exists() and not path_uvicorn2.exists():
            print("uvicorn not found in the Python path. Please install uvicorn using 'pip install uvicorn'.")
        else:
            if path_uvicorn1.exists():
                path_uvicorn = path_uvicorn1
            else:
                path_uvicorn = path_uvicorn2
                
    command = f"{path_uvicorn} main:app --port {port}"
    path = Path(__file__).parent
    os_type = platform.system()
    if os_type == "Darwin":
        server_process = subprocess.Popen(
            ['open', '-a', 'Terminal', path, '--args', '-c', command]
        )
    else:
        server_process = subprocess.Popen(['cmd', '/c', f'start', 'cmd', '/k', command], cwd=path)


def _start_server_in_same_terminal(port: int = 8000, new_thread=True) -> None:
    """Starts the server in a separate thread."""

    def run_server():
        """Internal function to configure and run the Uvicorn server."""
        upper_folder = Path(__file__).parent.resolve()
        sys.path.append(str(upper_folder))  # Ensure the path is absolute and resolved
        config = Config(app="main:app", port=port)  # Explicitly bind host
        server = Server(config)

        try:
            server.run()
        except Exception as e:
            print(f"Error: {e}")

    # Start the server in a new daemon thread
    if new_thread:
        server_thread = threading.Thread(target=run_server)
        server_thread.start()
        print("Server started on a separate thread.")

    else:
        run_server()


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--port", type=int, default=8000)
    parser.add_argument("--silent_browser", type=bool, default=False)
    args = parser.parse_args()

    
    os.chdir(os.path.dirname(__file__))
    if not args.silent_browser:
        webbrowser.open("http://localhost:8000/docs")
    _start_server_in_same_terminal(new_thread=False, port=args.port)
    # _start_server_in_new_terminal(port=args.port)
    



import os
import json
import tempfile
from pathlib import Path
from typing import Union
from AnyQt.QtWidgets import QLineEdit, QMessageBox
from Orange.data.table import Table
from Orange.widgets import widget, gui
from Orange.widgets.widget import Input
from Orange.widgets.settings import Setting
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.data.io import CSVReader


class OWDisplayOnInterface(widget.OWWidget):
    name = "Upload Statut"
    description = "Push statut to a local interface"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/local_interf_push.svg"

    priority = 1220
    category = "Advanced Artificial Intelligence Tools"
    want_main_area = False
    resizing_enabled = False

    class Inputs:
        data = Input("Data", Table)

    # Define an Error class for error messaging
    class Error(widget.OWWidget.Error):
        upload_failed = widget.Msg("Workflow ID does not match inputs. Upload failed.")

    # Persistent settings for fileId and workflow_id
    workflow_id: str = Setting('test.ows') # type: ignore
    messageStatut: str = Setting('') # type: ignore

    def __init__(self):
        super().__init__()
        self.info_label = gui.label(self.controlArea, self, "Initial info.")
        self.message_sta
        self.data = None
        self.base_folder = Path(__file__).parent.parent / "webserver" / "received_files"
        self.in_process_queue = self.base_folder / "in_process_queue.json"

        # Ensure the directory and queue file are present
        if not self.in_process_queue.exists():
            self.in_process_queue.parent.mkdir(parents=True, exist_ok=True)
            self.in_process_queue.touch(exist_ok=True)
        self.setup_ui()

    def get_output_folder(self) -> str:
        """Return the output folder for data saving."""
        try:
            with open(self.in_process_queue, 'r') as f:
                in_process_queue = json.load(f)
            
            for item in in_process_queue:
                if item['workflow_id'] == self.workflow_id:
                    return item['output_folder']
                
            raise ValueError("Workflow ID provided does not exist in the queue.")
        
        
        except ValueError:
            # Emit an error message when the workflow ID does not match
            self.Error.upload_failed()
            print("UploadWidget error. Workflow ID provided does not match the workflow id of inputs. Upload failed.")
            self.info_label.setText("UploadWidget error. Workflow ID provided does not match the workflow id of inputs. Upload failed.")
            raise ValueError("Workflow ID provided does not exist in the queue.")

    def setup_ui(self):
        """Set up the user interface."""

        # Text input for Workflow ID
        hbox2 = gui.hBox(self.controlArea, "Workflow ID")
        self.le_workflow_id = QLineEdit(self)
        self.le_workflow_id.setText(self.workflow_id)
        self.le_workflow_id.editingFinished.connect(self.update_workflow_id)
        hbox2.layout().addWidget(self.le_workflow_id) # type: ignore

        # Text input for statut to upload
        hbox3 = gui.hBox(self.controlArea, "MessageStatut to Upload")
        self.le_message_statut = QLineEdit(self)
        self.le_message_statut.setText(self.messageStatut)
        self.le_message_statut.editingFinished.connect(self.update_message_statut)
        hbox3.layout().addWidget(self.le_message_statut) # type: ignore

        self.adjustSize()


    def update_message_statut(self):
        """Update the message statut."""
        self.messageStatut = self.le_message_statut.text()

    @Inputs.data
    def dataset(self, data):
        """Handle new data input."""
        statut_path_str = Path(self.get_output_folder()) / "__statut__.txt"
            
    
    def update_workflow_id(self):
        """Update the workflow ID."""
        self.workflow_id = self.le_workflow_id.text()



if __name__ == "__main__": 
    pass
    # TODO: Implement a test for this widget.



import os
from pathlib import Path
from pprint import pprint

from AnyQt.QtWidgets import QMessageBox
from Orange.data import Table


if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.download.BaseLocalInterfaceWidget import BaseLocalInterfaceWidget
else:
    from orangecontrib.OWInterface.download.BaseLocalInterfaceWidget import BaseLocalInterfaceWidget



class OWLocalInterface(BaseLocalInterfaceWidget): # type: ignore
    name = "Local Interface - File"
    description = "Get a simple data file (csv, xlsx, pkl...) from a local interface"
    icon = "icons/local_interf_pull.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/local_interf_pull.svg"
    priority =1211
    category = "Advanced Artificial Intelligence Tools"


    def __init__(self):
        # Specific for this widget: defines how files are opened
        self.opening_method = "file"
        super().__init__()
        self.info_label.setText("Initialized Local Interface Widget.")

        

    def process_files(self, files_to_process: list[Path]):

        if len(files_to_process) > 1:
            print("More than one file to process. Only processing first. You may want to use MultiFile Download Widget")
            QMessageBox.critical(self, "DownloadWidget Error", "More than one file to process. Only processing first. You may want to use MultiFile Download Widget instead.")
        elif len(files_to_process) == 0:
            print("No files to process for Download Widget. Make sure you correctly set the input_id.")
            return
        file_path = files_to_process[0]

        try:
            # Load data using the defined opening method
            data_table = Table.from_file(str(file_path))
            self.info_label.setText(f"Loaded file: {str(file_path)}")
            self.Outputs.data.send(data_table)
        except Exception as e:
            print(f"Error loading file {str(file_path)}: {e}")
            self.info_label.setText(f"Error loading file: {str(file_path)}")
        return  # Only process first matching entry



import asyncio
import ctypes
import os
import threading
import time
from starlette.responses import JSONResponse


from AnyQt.QtWidgets import QFileDialog

dialog_should_close = threading.Event()

def set_dpi_awareness():
    if os.name == "nt":
        ctypes.windll.shcore.SetProcessDpiAwareness(1)

def open_file_dialog_sync():
    set_dpi_awareness()

    # Utilisation de QFileDialog.getOpenFileNames pour sélectionner plusieurs fichiers
    filepaths, _ = QFileDialog.getOpenFileNames(None, "Select files", "", "All Files (*)")

    if filepaths:
        print("Selected files:")
        for path in filepaths:
            print(path)
        return filepaths
    else:
        print("No file selected.")
        return ()

async def choose_file_on_server():
    loop = asyncio.get_event_loop()
    filepaths = await loop.run_in_executor(None, open_file_dialog_sync)
    return JSONResponse(content={"filepaths": filepaths})



import asyncio
import ctypes
import tkinter as tk
from tkinter import filedialog
from starlette.responses import JSONResponse
import threading
import time
import os
from typing import Literal

dialog_should_close = threading.Event()

def set_dpi_awareness():
    if os.name == "nt":
        ctypes.windll.shcore.SetProcessDpiAwareness(1)

def open_file_dialog_sync(mode: Literal["file", "folder"]):
    set_dpi_awareness()

    root = tk.Tk()
    root.iconify()
    
    if mode == "file":
        paths = filedialog.askopenfilenames()
    elif mode == "folder":
        paths = filedialog.askdirectory()
    root.destroy()

    if paths:
        return paths
    else:
        print("No file selected.")
        return ()

async def choose_file_on_server():
    loop = asyncio.get_event_loop()
    path = await loop.run_in_executor(None, open_file_dialog_sync, "file")
    return JSONResponse(content={"path": path})

async def choose_folder_on_server():
    loop = asyncio.get_event_loop()
    path = await loop.run_in_executor(None, open_file_dialog_sync, "folder")
    return JSONResponse(content={"path": path})


from typing import List, Literal, Union

from fastapi import FastAPI, File, HTTPException, UploadFile
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import FileResponse, JSONResponse, HTMLResponse

from functions.workflow_management import launch_workflow, stop_workflow, get_workflow_status 
from functions.file_upload import upload_files, upload_files_via_path
from functions.lifespan_manager import lifespan
from functions.output_retrieval import retrieve_outputs
from functions.fileDialog import choose_file_on_server, choose_folder_on_server
from settings import TIMEOUT_LIMIT, SERVER_DESCRIPTION, SERVER_NAME, html_examples_directory
from fastapi.staticfiles import StaticFiles
from pathlib import Path

app = FastAPI(
    title = SERVER_NAME,
    description = SERVER_DESCRIPTION,
    lifespan=lifespan)

app.mount("/html_examples", StaticFiles(directory=str(html_examples_directory), html=True), name="html_examples")


# CORS Settings
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],  # List of origins that are allowed, '*' means allow all origins
    allow_credentials=True,
    allow_methods=["*"],  # List of HTTP methods that are allowed, '*' means allow all methods
    allow_headers=["*"],  # List of headers that are allowed, '*' means allow all headers
    expose_headers=["Content-Disposition"]

)



# Main API Endpoints
@app.post("/upload-files/")
async def _upload_files(
    file_ids: List[str], 
    workflow_id: str,
    opening_methods: List[str], 
    expected_nb_outputs: int, 
    files: List[UploadFile] = File(default=None),
    text_inputs_id: List[str] = [],
    text_inputs_value: List[str] = [],
    timeout_limit: int = TIMEOUT_LIMIT,
) -> JSONResponse:
    """
    Upload multiple files to the server for processing through a Orange workflow. 
      
    Args:   
        - file_ids (List[str]): List of unique file IDs corresponding to the files.
          The IDs should match the ones configured in the Orange workflow.   
        - workflow_id (str): Unique ID of the workflow.
          The ID should match the one configured in the Orange workflow.   
        - opening_methods (List[str]): List of opening methods for the files.
          Choose from 'image_file', 'multiple_file', or 'file'.    
        - expected_nb_outputs (int): Expected number of output files (depending on your workflow).
        - files (List[UploadFile]): List of files to be uploaded.
        - text_inputs_id (List[str]): List of unique text input IDs corresponding to the text inputs.
            The IDs should match the ones configured in the Orange workflow.
        - text_inputs_value (List[str]): List of text input values corresponding to the text inputs.
        - timeout_limit (int): Timeout limit for the processing task (in seconds).

    Returns:   
        JSONResponse: Response message with the unique ID of the processing task."""

    return await upload_files(file_ids, workflow_id, opening_methods, expected_nb_outputs, files, text_inputs_id, text_inputs_value, timeout_limit)

@app.post("/upload-files-via-path/")
async def _upload_files_via_path(
    filepaths: List[str], 
    file_ids: List[str], 
    opening_methods: List[str],
    workflow_id: str, 
    expected_nb_outputs: int,
    text_inputs_id: List[str] = [],
    text_inputs_value: List[str] = [],
    timeout_limit: int = TIMEOUT_LIMIT,

) -> JSONResponse:
    """
    Upload multiple files to the server for processing through a Orange workflow.

    Args:   
        - filepaths (List[str]): List of file paths to be uploaded.   
        - file_ids (List[str]): List of unique file IDs corresponding to the files.
          The IDs should match the ones configured in the Orange workflow.  
        - opening_methods (List[str]): List of opening methods for the files.
          Choose from 'image_file', 'multiple_file', or 'file'.   
        - workflow_id (str): Unique ID of the workflow.
          The ID should match the one configured in the Orange workflow.   
        - expected_nb_outputs (int): Expected number of output files (depending on your workflow).
        - text_inputs_id (List[str]): List of unique text input IDs corresponding to the text inputs.
            The IDs should match the ones configured in the Orange workflow.
        - text_inputs_value (List[str]): List of text input values corresponding to the text inputs.
        - timeout_limit (int): Timeout limit for the processing task (in seconds).
    Returns:   
        JSONResponse: Response message with the unique ID of the processing task."""
    print("filepath received: ", filepaths)
    return await upload_files_via_path(filepaths, file_ids, opening_methods, workflow_id, expected_nb_outputs, text_inputs_id, text_inputs_value, timeout_limit)

@app.get("/retrieve-outputs/{unique_id}", response_model=None)
async def _retrieve_outputs(unique_id: str, 
                            file_id: str = "", 
                            return_mode: Literal["file", "json", "html"]="file",
                            delimiter = "\t") -> Union[FileResponse, JSONResponse, HTMLResponse]:
    """
    Retrieve the output files for a given unique ID. By default, a zip of all files will be returned.
    If a specific output filename is provided, only that file will be returned.   
    **IMPORTANT** : If the processing is not complete, the method will return a 404 error. You have to keep calling this method until the processing is complete.

    Args:     
        - unique_id (str): Unique ID of the processing task.  
        - file_id (str): Optional file ID to retrieve a specific file. If not provided, a zip of all files will be returned. 

    Returns:  
        FileResponse ou JSONResponse: Response with the output file(s) for download.
    """
    return await retrieve_outputs(unique_id, file_id, return_mode, delimiter)


@app.post("/launch-workflow/")
async def _launch_workflow(
    workflow_name: str,
    force_reload: bool = False,
    gui: bool = True
) -> JSONResponse:
    """
    Launch a workflow from an Orange file. The workflow will be executed in the background.

    Args:   
        - workflow (File): Orange workflow file to be executed.
    
    Returns:   
        JSONResponse: Response message, indicating that the workflow has been launched."""
    return await launch_workflow(workflow_name, force_reload, gui)
    
    
@app.post("/stop-workflow/")
async def _stop_workflow(
    workflow_id: str
) -> JSONResponse:
    """
    Stop a workflow that is currently running.

    Args:   
        - workflow_id (str): Unique ID of the workflow to be stopped.
    
    Returns:   
        JSONResponse: Response message, indicating that the workflow has been stopped."""
    return await stop_workflow(workflow_id)

@app.get("/get-running-workflows/")
async def _get_running_workflows() -> JSONResponse:
    """
    Get a list of all running workflows.

    Returns:   
        JSONResponse: Response message with the list of running workflows."""
    return await get_workflow_status()


@app.get("/choose-file-on-server/")
async def _choose_file_on_server() -> JSONResponse:
    """Open a file dialog to choose a file on the server."""
    return await choose_file_on_server()


@app.get("/select-folder/")
async def _select_folder() -> JSONResponse:
    """
    Opens a dialog window to let the user select a folder path.
    """
    return await choose_folder_on_server()


